import typing as t
from numbers import Number
from .base import BaseSchema


class NumberSchema(BaseSchema):
    def __init__(self):
        super().__init__((int, float))

    def float(self):
        self.type = float
        return self

    def int(self):
        self.type = int
        return self

    def lt(self, num: t.Type[Number]):
        return self.add_validator(
            'is-lt',
            lambda data: None if data < num else f'number must be less than {num}',
            False
        )

    def gt(self, num: t.Type[Number]):
        return self.add_validator(
            'is-gt',
            lambda data: None if data > num else f'number must be greater than {num}',
            False
        )

    def lte(self, num: t.Type[Number]):
        return self.add_validator(
            'is-lte',
            lambda data: None if data <= num else f'number must be less than or equal to {num}',
            False
        )

    def gte(self, num: t.Type[Number]):
        return self.add_validator(
            'is-gte',
            lambda data: None if data >= num else f'number must be greater than or equal to {num}',
            False
        )

    def min(self, num: t.Type[Number]):
        return self.gte(num)

    def max(self, num: t.Type[Number]):
        return self.lte(num)

    def positive(self):
        return self.gt(0)

    def negative(self):
        return self.lt(0)



import typing as t
from dataclasses import dataclass


@dataclass
class Validator:
    name: str
    fn: t.Callable[
        [
            t.Any,
            # t.Type['BaseSchema']
        ],
        t.Optional[str]
    ]

    def __call__(self, data: t.Any):
        return self.fn(data)
    

class BaseSchema:
    def __init__(self, type: t.Union[t.Tuple[t.Callable], t.Callable], type_validator:t.Optional[t.Callable]=None):
        self.type = type
        
        if not type_validator:
            def type_validator(data):
                if not isinstance(data, self.type):
                    if isinstance(self.type, tuple):
                        t_str = ', '.join(map(lambda t: t.__name__, self.type))
                        return f'data type is not of expected types: {t_str}'
                    return f'{data} is not of type {self.type.__name__}'
                    
        self._validators: t.List['Validator'] = [Validator('is-type', type_validator)]
        self._throw: bool = False
        self._errors: t.List[str] = []

    def __repr__(self):
        return f'<{self.__class__.__name__} of kohi>'

    __str__ = __repr__

    @property
    def errors(self):
        return self._errors

    def _handle_errors(self):
        if not self._throw or len(self._errors) < 1:
            return
        
        message = self._errors[0]
        
        if len(self._errors) > 1:
            message = f'{len(self._errors)} errors occurred'
            
        ValidationError = type('ValidationError', (Exception,), {})
        raise ValidationError(message)

    def add_validator(self, name: str, fn: t.Callable[[t.Any], t.Optional[str]], exclusive:bool=True):
        if exclusive and any(map(lambda el: el.name == name, self._validators)):
            return
        self._validators.append(Validator(name=name, fn=fn))
        return self

    def validate(self, data: t.Any):
        """Validate the given data against the schema"""
        self.reset()
        for validator in self._validators:
            error = validator(data)
            if error:
                self._errors.append(error)
        self._handle_errors()                
        return len(self.errors) == 0

    def reset(self):
        self._errors = []
    
    def throw(self):
        self._throw = True
        return self
        


import re, uuid
from .base import BaseSchema

__all__ = ('String',)


class StringSchema(BaseSchema):
    def __init__(self):
        super().__init__((str,))

    def min(self, min_length: int):
        return self.add_validator(
            'min-length',
            lambda data: None if len(data) >= min_length else f'data length must be greater than or equal to {min_length}'
        )
    
    def length(self, length: int):
        return self.add_validator(
            'length',
            lambda data: None if len(data) == length else f'data length must be equal to {length}'
        )

    def max(self, max_length: int):
        return self.add_validator(
            'max-length',
            lambda data: None if len(data) <= max_length else f'data length must be less than or equal to {max_length}'
        )
    
    def url(self):
        def test(data):
            regex = re.compile(
                r'^(?:http|ftp)s?://'  # http:// or https://
                r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
                r'localhost|'  # localhost...
                r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
                r'(?::\d+)?'  # optional port
                r'(?:/?|[/?]\S+)$', re.IGNORECASE)
            
            if not regex.match(data):
                return 'string is not a valid URL'
        return self.add_validator('is-url', test)

    def uuid(self):
        def test(data):
            try:
                uuid.UUID(data)
            except ValueError:
                return 'string is not a valid UUID'
        return self.add_validator('is-uuid', test)

    def starts_with(self, text: str):
        return self.add_validator('starts-with', lambda data: None if data.startswith(text) else f'string must starts with {text}')

    def ends_with(self, text: str):
        return self.add_validator('ends-with', lambda data: None if data.endswith(text) else f'string must end with {text}')

    # def email(self):
    #     ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kohi']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'kohi',
    'version': '0.1.0',
    'description': 'A powerfull schema validator',
    'long_description': '\n# kohi\n\n<p align="center">A powerfull schema validator</p>\n\n![GitHub Repo stars](https://img.shields.io/github/stars/natanfeitosa/kohi)\n![GitHub Workflow Status](https://img.shields.io/github/actions/workflow/status/natanfeitosa/kohi/pytest.yml?label=Pytest&logo=github)\n![GitHub](https://img.shields.io/github/license/natanfeitosa/kohi)\n\n## Quickstart\n\nTo validate a type you can import your schema validator from `kohi` or `from kohi.<type> import <type>Schema`\n\ne.g.\n\nLet\'s check if a person\'s date of birth is a positive integer less than the current date — 2023 — and greater than or equal to 2005\n\n```python\nfrom kohi import NumberSchema\n# from kohi.number import NumberSchema\n\nn = NumberSchema().int().positive().lt(2023).gte(2005)\n\nprint(n.validate(2005)) # True\nprint(n.validate(2022)) # True\nprint(n.validate(2004)) # False\nprint(n.validate(2023)) # False\n\n# You can see the errors generated in the last `validate` call just by accessing the `errors` property\n# print(n.errors) # [\'number must be less than 2022\']\n```\n\n## Validators\n\n* [`kohi.base.BaseSchema`](#baseschema)\n> Only one base class for all schema validators\n* [`kohi.number.NumberSchema`](#numberschema)\n> or `kohi.NumberSchema`\n* [`kohi.string.StringSchema`](#stringschema)\n> or `kohi.StringSchema`\n\n## Methods\n\n### `BaseSchema`\n* `add_validator(name, func): Self`\n  > Add a custom data validator\n* `validate(data): bool`\n  > The method to be called when we validate the schema\n* `reset(): None`\n  > Reset error list\n* `throw(): Self`\n  > By default no errors are thrown, but when this method is chained a `ValidationError` will be thrown\n\n### `NumberSchema`\ninherits from [`BaseSchema`](#baseschema)\n> By default validates int and float \n\n* `float(): Self`\n  > Validate only `float`\n* `int(): Self`\n  > Validate only `int`\n* `lt(num): Self`\n  > Validates if the data is less than `num`\n* `gt(num): Self`\n  > Validates if the data is greater than `num`\n* `lte(num): Self`\n  > Validates if the data is less than or equal to `num`\n* `gte(num): Self`\n  > Validates if the data is greater than or equal to `num`\n* `min(num): Self`\n  > Just an alias for `gte(num)`\n* `max(num): Self`\n  > Just an alias for `lte(nun)`\n* `positive(): Self`\n  > Just an alias for `gt(0)`\n* `negative(): Self`\n  > Just an alias for `lt(0)`\n\n### StringSchema\ninherits from [`BaseSchema`](#baseschema)\n\n## Dev env\n\n* install development dependencies\n* check types using `mypy`\n* run all tests using `pytest`\n',
    'author': 'Natan Santos',
    'author_email': 'natansantosapps@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/natanfeitosa/kohi#readme',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
